/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.plugins.sql

import groovy.sql.Sql
import groovy.text.SimpleTemplateEngine
import javax.sql.DataSource
import metridoc.utils.Assert
import metridoc.utils.PropertyUtils
import org.apache.commons.lang.StringEscapeUtils
import org.apache.commons.lang.math.NumberUtils
import org.slf4j.LoggerFactory
import java.util.regex.Pattern
import java.util.regex.Matcher
import java.sql.Statement
import java.sql.BatchUpdateException
import groovy.sql.BatchingStatementWrapper
import metridoc.utils.SystemUtils
import java.awt.print.PageFormat

/**
 * Created by IntelliJ IDEA.
 * User: tbarker
 * Date: 7/18/11
 * Time: 3:38 PM
 */
class SqlPlus extends Sql {

    static final PHASE_NAMES = "phaseName"
    static final log = LoggerFactory.getLogger(SqlPlus.class)
    static final Set IGNORED_KEYS = [PHASE_NAMES, "order"]

    SqlPlus(DataSource dataSource) {
        super(dataSource)
    }

    void runBatchFile(LinkedHashMap args) {
        Assert.notNull(args.fileName, "SqlPlus requires a fileName to runBatchFile")
        ConfigObject file = new PropertyUtils().getConfig(args.fileName)
        def phases = getPhases(file)
        phases.each {order, phase ->
            long beginPhaseTime = new Date().getTime()
            def phaseName = phase.phaseName
            log.info "starting phase ${phaseName}"
            if (!exclude(args.exclude, phaseName)) {
                phase.each {sqlName, value ->
                    def fullSqlName = "${phaseName}.${sqlName}"
                    if (!exclude(args.exclude, fullSqlName)) {

                        if (!IGNORED_KEYS.contains(sqlName)) {
                            log.info("running ${fullSqlName}")
                            long startTime = new Date().getTime()
                            int updateCount = executeUpdate(value.sql)
                            long endTime = new Date().getTime()
                            long totalTime = endTime - startTime
                            log.info("finished running ${fullSqlName} with ${updateCount} updates, took ${totalTime} milliseconds")
                        }
                    } else {
                        log.info("skipping sql ${fullSqlName}")
                    }

                }
            } else {
                log.info("skipping phase ${phaseName}")
            }

            long endPhaseTime = new Date().getTime()
            long totalPhaseTime = endPhaseTime - beginPhaseTime
            log.info "finished ${phaseName}, took ${totalPhaseTime} milliseconds"
        }
    }

    private static boolean exclude(exclude, String value) {
        if (exclude) {
            return exclude.contains(value)
        }

        return false
    }

    private static SortedMap getPhases(ConfigObject configObject) {
        def result = new TreeMap()
        int defaultOrder = 1000
        ConfigObject sqlPhases = configObject.sql
        sqlPhases.each {key, value ->
            try {
                value[PHASE_NAMES] = key
                value.order = value.containsKey("order") ? value.order : defaultOrder
                result.put(Double.valueOf(value.order.toString()), value)
            } catch (Exception ex) {
                throw new RuntimeException("there was an error loading phase ${key}", ex)
            }
        }

        return result
    }

    /**
     *
     * @param insertOrTable insert statement or table name
     * @param batch the batch to insert, must be a {@link List} or {@link Map}
     * @param logEachBatch if true, batch progress is logged at info level, otherwise debug
     * @return an array of integers that indicate the number of updates for each statement
     */
    int[] runBatch(String insertOrTable, batch, logEachBatch = false) {
        try {
            int[] result = withBatch {statement ->
                if (batch instanceof List) {
                    runListBatch(batch, insertOrTable, statement)
                } else {
                    runMapBatch(insertOrTable, batch, statement)
                }
            }

            logBatch(result, logEachBatch)
            return result
        } catch (BatchUpdateException ex) {
            if (batch instanceof List) {
                def records = batch as List
                int[] updates = ex.updateCounts
                int upperBound = updates.size() - 1

                (0..upperBound).each {int index ->
                    int update = updates[index]
                    if (failed(update)) {
                        logFailedRecord(records[index])
                    }
                }

                if(updates.size() < records.size()) {
                    log.error("some exception prevented execution of any records added after and including {}",
                            records[updates.size()])
                }
            }

            if(batch instanceof Map) {
                logFailedRecord(batch)
            }

            throw ex //now that we logged what we can, throw it
        }
    }

    private static logBatch(int[] result, boolean logEachBatch) {
        if(shouldLog(logEachBatch)) {
            int recordCount = result.size()
            int totalUpdates = 0
            result.each{
                totalUpdates += it
            }
            String message = "processed ${recordCount} records with ${totalUpdates} updates"
            if(logEachBatch) {
                log.info(message)
            } else {
                log.debug(message)
            }
        }
    }

    private static boolean shouldLog(boolean logEachBatch) {
        if(logEachBatch) {
            return log.isInfoEnabled()
        }

        return log.isDebugEnabled()
    }

    private static void logFailedRecord(record) {
        log.error("record {} failed to be executed in batch", record)
    }

    private static boolean failed(int update) {
        return update == Statement.EXECUTE_FAILED
    }

    private def runMapBatch(String insertOrTable, Map batch, BatchingStatementWrapper statement) {
        def insert = getInsertStatement(insertOrTable, batch)
        if (batch == null) {
            throw new IllegalArgumentException("a record must be a none null Map to use batch inserting")
        }

        if (!(batch instanceof Map)) {
            throw new IllegalArgumentException("record ${batch} must be of type Map to use batch inserting")
        }

        logRecordInsert(insert, batch)
        statement.addBatch(insert)
    }

    private runListBatch(List batch, String insertOrTable, BatchingStatementWrapper statement) {
        batch.each {record ->

            if (record == null) {
                throw new IllegalArgumentException("a record must be a none null Map to use batch inserting")
            }

            if (!(record instanceof Map)) {
                throw new IllegalArgumentException("record ${record} must be of type Map to use batch inserting")
            }

            def insert = getInsertStatement(insertOrTable, record)
            logRecordInsert(insert, record)
            statement.addBatch(insert)
        }

        log.debug("finished adding {} records to batch, now the batch will be executed", batch.size())
    }

    private static void logRecordInsert(String insert, record) {
        log.debug("adding {} to batch inserts, which was created from {}", insert, record)
    }

    private static String getInsertStatement(String tableOrInsert, Map values) {
        def words = tableOrInsert.split()

        //must be an update statement of some sort (insert, update, etc.)
        if (words.size() > 1) {
            return getInsertStatementFromParamInsert(tableOrInsert, values)
        }

        return getInsertStatementForTable(tableOrInsert, values)
    }

    private static getInsertStatementForTable(String table, Map values) {
        log.debug("retrieving insert statement for table {} using record {}", table, values)
        StringBuffer insert = new StringBuffer("insert into ")
        StringBuffer valuesToInsert = new StringBuffer("values (")
        insert.append(table)
        insert.append(" (")
        values.each {key, value ->
            insert.append(key)
            insert.append(", ")
            String valueText = String.valueOf(value)

            if (NumberUtils.isNumber(valueText)) {
                valuesToInsert.append(value)
            } else {
                valuesToInsert.append("'")
                valuesToInsert.append(StringEscapeUtils.escapeSql(value))
                valuesToInsert.append("'")
            }

            valuesToInsert.append(", ")
        }
        insert.delete(insert.length() - 2, insert.length())
        insert.append(") ")
        valuesToInsert.delete(valuesToInsert.length() - 2, valuesToInsert.length())
        valuesToInsert.append(")")

        return insert.append(valuesToInsert).toString()
    }

    private static String getInsertStatementFromParamInsert(String insert, Map values) {
        def formattedInsert = formatParamInsert(insert)
        def engine = new SimpleTemplateEngine()

        //need to have a separate binding since extra values are added by the template engine
        def binding = [:]
        binding.putAll(values)
        return engine.createTemplate(formattedInsert).make(binding).toString()
    }

    private static String formatParamInsert(String insert) {

        def varTags = [':', '#']
        def result = insert

        varTags.each {String tag ->
            result = regexReplace(tag, result, false)
        }

        result = regexReplace('\\$', result, true)

        return result
    }

    private static String regexReplace(String varPrefix, String text, boolean addQuotes) {
        StringBuilder patternToSearch = new StringBuilder()
        patternToSearch.append('([^\'])')
        patternToSearch.append(varPrefix)
        patternToSearch.append('(\\w+)')

        Pattern p = Pattern.compile(patternToSearch.toString())
        def result = text
        Matcher m = p.matcher(result)

        while (m.find()) {
            def replacement = new StringBuilder()
            replacement.append(m.group(1))
            addQuote(replacement, addQuotes)

            replacement.append('\\$')
            replacement.append(m.group(2))

            addQuote(replacement, addQuotes)

            result = m.replaceFirst(replacement.toString())
            m = p.matcher(result)
        }

        return result
    }

    private static addQuote(StringBuilder builder, boolean addQuote) {
        if (addQuote) {
            builder.append('\'')
        }
    }
}

